@@ -6,6 +6,7 @@
   Copyright 2008      Y Giridhar Appaji Nag
   Copyright 2016      Ivan Gimenez
   Copyright 2016      Stephen Thirlwall
+  Copyright 2017      Antonio Quartulli
   Copyright 2017      Ismael Luceno
 
   This program is free software; you can redistribute it and/or
@@ -42,26 +43,52 @@
 #include "axel.h"
 
 /* Some nifty macro's.. */
-#define get_config_string( name )				\
-	if( strcmp( key, #name ) == 0 )				\
-	{							\
-		st = 1;						\
-		strcpy( conf->name, value );			\
-	}
-#define get_config_number( name )				\
-	if( strcmp( key, #name ) == 0 )				\
-	{							\
-		st = 1;						\
-		sscanf( value, "%i", &conf->name );		\
-	}
+#define MATCH if( 0 );
+#define KEY( name )				\
+	else if( !strcmp( key, #name ) )	\
+		dst = &conf->name;
 
 int parse_interfaces( conf_t *conf, char *s );
 
+static int axel_fscanf( FILE *fp, const char *format, ...)
+{
+	va_list params;
+	int ret;
+
+	va_start( params, format );
+	ret = vfscanf( fp, format, params );
+	va_end( params );
+
+	ret = !( ret == EOF  && ferror(fp) );
+	if( !ret )
+	{
+		fprintf( stderr, _("I/O error while reading config file: %s\n"),
+			 strerror(errno) );
+	}
+
+	return( ret );
+}
+
+static int parse_protocol( conf_t *conf, const char *value )
+{
+	if( strcasecmp( value, "ipv4" ) == 0 )
+		conf->ai_family = AF_INET;
+	else if( strcasecmp( value, "ipv6" ) == 0 )
+		conf->ai_family = AF_INET6;
+	else
+	{
+		fprintf( stderr, _( "Unknown protocol %s\n" ), value );
+		return( 0 );
+	}
+
+	return( 1 );
+}
+
 int conf_loadfile( conf_t *conf, char *file )
 {
-	int line = 0;
+	int line = 0, ret = 1;
 	FILE *fp;
-	char s[MAX_STRING], key[MAX_STRING], value[MAX_STRING];
+	char s[MAX_STRING], key[MAX_STRING];
 
 	fp = fopen( file, "r" );
 	if( fp == NULL )
@@ -69,75 +96,97 @@ int conf_loadfile( conf_t *conf, char *file )
 
 	while( !feof( fp ) )
 	{
-		int st;
+		char *tmp, *value = NULL;
+		void *dst;
 
 		line ++;
 
 		*s = 0;
-		fscanf( fp, "%100[^\n#]s", s );
-		fscanf( fp, "%*[^\n]s" );
+
+		if( !( ret = axel_fscanf( fp, "%100[^\n#]s", s ) ) )
+			break;
+		if( !( ret = axel_fscanf( fp, "%*[^\n]s" ) ) )
+			break;
 		fgetc( fp );			/* Skip newline */
-		if( strchr( s, '=' ) == NULL )
+		tmp = strchr( s, '=' );
+		if( tmp == NULL )
 			continue;		/* Probably empty? */
+
 		sscanf( s, "%[^= \t]s", key );
-		{
-			int i;
-			for( i = 0; s[i]; i ++ )
-				if( s[i] == '=' )
-				{
-					for( i ++; isspace( (int) s[i] ) && s[i]; i ++ );
-					break;
-				}
-			strcpy( value, &s[i] );
-		}
-		/* FIXME: could be optimized to re-use the previous loop and
-		   write a single zero. */
-		for( int i = strlen( value ) - 1; isspace( (int) value[i] ); i -- )
-			value[i] = 0;
 
-		st = 0;
+		/* Skip the "=" and any spaces following it */
+		while( isspace( * ++ tmp ) ); /* XXX isspace('\0') is false */
+		value = tmp;
+		/* Get to the end of the value string */
+		while ( *tmp && !isspace( *tmp ) ) tmp++;
+		*tmp = '\0';
 
-		/* Long live macros!! */
-		get_config_string( default_filename );
-		get_config_string( http_proxy );
-		get_config_string( no_proxy );
-		get_config_number( strip_cgi_parameters );
-		get_config_number( save_state_interval );
-		get_config_number( connection_timeout );
-		get_config_number( reconnect_delay );
-		get_config_number( num_connections );
-		get_config_number( buffer_size );
-		get_config_number( max_speed );
-		get_config_number( verbose );
-		get_config_number( alternate_output );
-		get_config_number( insecure );
+		/* String options */
+		MATCH
+			KEY( default_filename )
+			KEY( http_proxy )
+			KEY( no_proxy )
+			KEY( user_agent )
+		else goto num_keys;
 
-		get_config_number( search_timeout );
-		get_config_number( search_threads );
-		get_config_number( search_amount );
-		get_config_number( search_top );
+		/* Save string option */
+		strcpy( dst, value );
+		continue;
 
+		/* Numeric options */
+	num_keys:
+		MATCH
+			KEY( strip_cgi_parameters )
+			KEY( save_state_interval )
+			KEY( connection_timeout )
+			KEY( reconnect_delay )
+			KEY( num_connections )
+			KEY( max_redirect )
+			KEY( buffer_size )
+			KEY( max_speed )
+			KEY( verbose )
+			KEY( alternate_output )
+			KEY( insecure )
+			KEY( no_clobber )
+			KEY( search_timeout )
+			KEY( search_threads )
+			KEY( search_amount )
+			KEY( search_top )
+		else goto other_keys;
+
+		/* Save numeric option */
+		*( ( int * ) dst ) = atoi( value );
+		continue;
+
+	other_keys:
 		/* Option defunct but shouldn't be an error */
 		if( strcmp( key, "speed_type" ) == 0 )
-			st = 1;
-
-		if( strcmp( key, "interfaces" ) == 0 )
-			st = parse_interfaces( conf, value );
-
-		if( !st )
+			continue;
+		else if( strcmp( key, "interfaces" ) == 0 )
 		{
-			fprintf( stderr, _("Error in %s line %i.\n"), file, line );
-			fclose( fp );
-			return( 0 );
+			if( parse_interfaces( conf, value ) )
+				continue;
 		}
+		else if( strcmp( key, "use_protocol" ) == 0 )
+		{
+			if( parse_protocol( conf, value ) )
+				continue;
+		}
+
+#if 0
+		/* FIXME broken code */
 		get_config_number( add_header_count );
 		for(int i=0;i<conf->add_header_count;i++)
 			get_config_string( add_header[i] );
-		get_config_string( user_agent );
+#endif
+
+		fprintf( stderr, _("Error in %s line %i.\n"), file, line );
+		ret = 0;
+		break;
 	}
 
 	fclose( fp );
-	return( 1 );
+	return( ret );
 }
 
 int conf_init( conf_t *conf )
@@ -155,17 +204,22 @@ int conf_init( conf_t *conf )
 	conf->connection_timeout	= 45;
 	conf->reconnect_delay		= 20;
 	conf->num_connections		= 4;
+	conf->max_redirect		= MAX_REDIRECT;
 	conf->buffer_size		= 5120;
 	conf->max_speed			= 0;
 	conf->verbose			= 1;
 	conf->alternate_output		= 0;
 	conf->insecure			= 0;
+	conf->no_clobber		= 0;
 
 	conf->search_timeout		= 10;
 	conf->search_threads		= 3;
 	conf->search_amount		= 15;
 	conf->search_top		= 3;
 	conf->add_header_count		= 0;
+
+	conf->ai_family			= AF_UNSPEC;
+
 	strncpy( conf->user_agent, DEFAULT_USER_AGENT, MAX_STRING );
 
 	conf->interfaces = malloc( sizeof( if_t ) );
@@ -197,6 +251,12 @@ int conf_init( conf_t *conf )
 	return( 1 );
 }
 
+/* release resources allocated by conf_init() */
+void conf_free( conf_t *conf )
+{
+	free( conf->interfaces );
+}
+
 int parse_interfaces( conf_t *conf, char *s )
 {
 	char *s2;
